from attr import validators, NOTHING
from .validators import composite


def init_default(required, default, optional_default):
    """
    Returns optional default if field is not required and
    default was not provided.

    :param bool required: whether the field is required in a given model.
    :param default: default provided by creator of field.
    :param optional_default: default for the data type if none provided.
    :return: default or optional default based on inputs
    """
    if not required and default == NOTHING:
        default = optional_default

    return default


def init_validator(required, cls, *additional_validators):
    """
    Create an attrs validator based on the cls provided and required setting.
    :param bool required: whether the field is required in a given model.
    :param cls: the expected class type of object value.
    :return: attrs validator chained correctly (e.g. optional(instance_of))
    """
    validator = validators.instance_of(cls)

    if additional_validators:
        additional_validators = list(additional_validators)
        additional_validators.append(validator)
        validator = composite(*additional_validators)

    return validator if required else validators.optional(validator)



# -*- coding: utf-8 -*-
from attr.exceptions import FrozenInstanceError
from collections import OrderedDict

try:
    from collections.abc import (MutableSequence, MutableMapping, MutableSet)
except ImportError:
    from collections import (MutableSequence, MutableMapping, MutableSet)


DEFAULT_DATE_FORMAT = "%Y-%m-%d"
DEFAULT_DATETIME_FORMAT = "ISO_FORMAT"
DEFAULT_TIME_FORMAT = "%H:%M:%S"


class ImmutableDict(dict):

    def __setitem__(self, key, value):
        raise FrozenInstanceError()

    def __delitem__(self, key):
        raise FrozenInstanceError()

    def __setattr__(self, name, value):
        raise FrozenInstanceError()

    def __delattr__(self, name):
        raise FrozenInstanceError()

    def pop(self, key, **kwargs):
        raise FrozenInstanceError()

    def clear(self):
        raise FrozenInstanceError()


class TypedSequence(MutableSequence):
    """
    Custom list type that checks the instance type of new values.

    reference:
    http://stackoverflow.com/a/3488283
    """

    def __init__(self, cls, args, allow_none=True):
        self.cls = cls
        self.allowed_types = (cls, type(None)) if allow_none else cls
        self.list = []
        self.extend(args)

    def __str__(self):
        return str(self.list)

    def __repr__(self):
        return repr(self.list)

    def __len__(self):
        return len(self.list)

    def __eq__(self, other):
        if isinstance(other, TypedSequence):
            return self.list == other.list and self.cls == other.cls
        else:
            return self.list == other

    def __getitem__(self, i):
        return self.list[i]

    def __delitem__(self, i):
        del self.list[i]

    def __setitem__(self, i, v):
        self._check(v)
        self.list[i] = v

    def insert(self, i, v):
        self._check(v)
        self.list.insert(i, v)

    def _check(self, v):
        if not isinstance(v, self.allowed_types):
            raise TypeError("Invalid value %s (%s != %s)" %
                            (v, type(v), self.cls))


class TypedMapping(MutableMapping):
    """
    Custom dict type that checks the instance type of new values.

    reference:
    http://stackoverflow.com/a/3488283
    """

    def __init__(self, cls, kwargs, key=None, allow_none=True):
        self.cls = cls
        self.allowed_types = (cls, type(None)) if allow_none else cls
        self.key = key
        self.dict = OrderedDict()
        self.update(kwargs)

    def __str__(self):
        return str(self.dict)

    def __repr__(self):
        return repr(self.dict)

    def __len__(self):
        return len(self.dict)

    def __iter__(self):
        return iter(self.dict)

    def __eq__(self, other):
        if isinstance(other, TypedMapping):
            return self.dict == other.dict and self.cls == other.cls
        else:
            return self.dict == other

    def __getitem__(self, i):
        return self.dict[i]

    def __delitem__(self, i):
        del self.dict[i]

    def __setitem__(self, i, v):
        self._check(v)
        self.dict[i] = v

    def add(self, v, key=None):
        if key is not None:
            self[key] = v

        elif self.key and hasattr(v, self.key):
            self[getattr(v, self.key)] = v  # pragma: no cover

        else:
            raise TypeError("No key set in instance or provided in call.")

    def _check(self, v):
        if not isinstance(v, self.allowed_types):
            raise TypeError("%s is not an instance of %s" % (v, self.cls))


class TypedSet(MutableSet):
    """
    Custom set type that checks the instance type of new values.

    reference:
    http://stackoverflow.com/a/3488283
    """

    def __init__(self, cls, args, allow_none=True):
        self.cls = cls
        self.allowed_types = (cls, type(None)) if allow_none else cls
        self.set = set()
        for arg in args or []:
            self.add(arg)

    def __str__(self):
        return str(self.set)

    def __repr__(self):
        return repr(self.set)

    def __len__(self):
        return len(self.set)

    def __eq__(self, other):
        if isinstance(other, TypedSet):
            return self.set == other.set and self.cls == other.cls
        else:
            return self.set == other

    def __iter__(self):
        return iter(self.set)

    def __contains__(self, item):
        return item in self.set

    def add(self, v):
        self._check(v)
        self.set.add(v)

    def discard(self, value):
        self.set.discard(value)

    def _check(self, v):
        if not isinstance(v, self.allowed_types):
            raise TypeError("Invalid value %s (%s != %s)" %
                            (v, type(v), self.cls))



from __future__ import absolute_import, division, print_function

from collections import OrderedDict
from enum import Enum

import yaml
import json

from attr._make import fields

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch


@singledispatch
def to_dict(obj, **kwargs):
    """
    Convert an object into dictionary. Uses singledispatch to allow for
    clean extensions for custom class types.

    Reference: https://pypi.python.org/pypi/singledispatch

    :param obj: object instance
    :param kwargs: keyword arguments such as suppress_private_attr,
                   suppress_empty_values, dict_factory
    :return: converted dictionary.
    """

    # if is_related, then iterate attrs.
    if is_model(obj.__class__):
        return related_obj_to_dict(obj, **kwargs)

    # else, return obj directly. register a custom to_dict if you need to!
    #   reference: https://pypi.python.org/pypi/singledispatch
    else:
        return obj


def related_obj_to_dict(obj, **kwargs):
    """ Covert a known related object to a dictionary. """

    # Explicitly discard formatter kwarg, should not be cascaded down.
    kwargs.pop('formatter', None)

    # If True, remove fields that start with an underscore (e.g. _secret)
    suppress_private_attr = kwargs.get("suppress_private_attr", False)

    # if True, don't store fields with None values into dictionary.
    suppress_empty_values = kwargs.get("suppress_empty_values", False)

    # get list of attrs fields
    attrs = fields(obj.__class__)

    # instantiate return dict, use OrderedDict type by default
    return_dict = kwargs.get("dict_factory", OrderedDict)()

    for a in attrs:

        # skip if private attr and flag tells you to skip
        if suppress_private_attr and a.name.startswith("_"):
            continue

        metadata = a.metadata or {}

        # formatter is a related-specific `attrs` meta field
        #   see fields.DateField
        formatter = metadata.get('formatter')

        # get value and call to_dict on it, passing the kwargs/formatter
        value = getattr(obj, a.name)
        value = to_dict(value, formatter=formatter, **kwargs)

        # check flag, skip None values
        if suppress_empty_values and value is None:
            continue

        # field name can be overridden by the metadata field
        key_name = a.metadata.get('key') or a.name

        # store converted / formatted value into return dictionary
        return_dict[key_name] = value

    return return_dict


def to_model(cls, value):
    """
    Coerce a value into a model object based on a class-type (cls).
    :param cls: class type to coerce into
    :param value: value to be coerced
    :return: original value or coerced value (value')
    """

    if isinstance(value, cls) or value is None:
        pass  # skip if right type or value is None

    elif issubclass(cls, Enum):
        value = cls(value)

    elif is_model(cls) and isinstance(value, dict):
        value = convert_key_to_attr_names(cls, value)
        value = cls(**value)

    else:
        value = cls(value)

    return value


def convert_key_to_attr_names(cls, original):
    """ convert key names to their corresponding attribute names """
    attrs = fields(cls)
    updated = {}
    keys_pulled = set()

    for a in attrs:
        key_name = a.metadata.get('key') or a.name
        if key_name in original:
            updated[a.name] = original.get(key_name)
            keys_pulled.add(key_name)

    if getattr(cls, '__related_strict__', False):
        extra = set(original.keys()) - keys_pulled
        if len(extra):
            raise ValueError("Extra keys (strict mode): {}".format(extra))

    return updated


def is_model(cls):
    """
    Check whether *cls* is a class with ``attrs`` attributes.

    :param type cls: Class to introspect.
    :raise TypeError: If *cls* is not a class.

    :rtype: :class:`bool`
    """
    return getattr(cls, "__attrs_attrs__", None) is not None


def to_yaml(obj, stream=None, dumper_cls=yaml.Dumper, default_flow_style=False,
            **kwargs):
    """
    Serialize a Python object into a YAML stream with OrderedDict and
    default_flow_style defaulted to False.

    If stream is None, return the produced string instead.

    OrderedDict reference: http://stackoverflow.com/a/21912744
    default_flow_style reference: http://stackoverflow.com/a/18210750

    :param data: python object to be serialized
    :param stream: to be serialized to
    :param Dumper: base Dumper class to extend.
    :param kwargs: arguments to pass to to_dict
    :return: stream if provided, string if stream is None
    """

    class OrderedDumper(dumper_cls):
        pass

    def dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            data.items())

    OrderedDumper.add_representer(OrderedDict, dict_representer)

    obj_dict = to_dict(obj, **kwargs)

    return yaml.dump(obj_dict, stream, OrderedDumper,
                     default_flow_style=default_flow_style)


def from_yaml(stream, cls=None, loader_cls=yaml.Loader,
              object_pairs_hook=OrderedDict, **extras):
    """
    Convert a YAML stream into a class via the OrderedLoader class.
    """

    class OrderedLoader(loader_cls):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    yaml_dict = yaml.load(stream, OrderedLoader) or {}
    yaml_dict.update(extras)
    return cls(**yaml_dict) if cls else yaml_dict


def to_json(obj, indent=4, sort_keys=True, **kwargs):
    """
    :param obj: object to convert to dictionary and then output to json
    :param indent: indent json by number of spaces
    :param sort_keys: sort json output by key if true
    :param kwargs: arguments to pass to to_dict
    :return: json string
    """
    obj_dict = to_dict(obj, **kwargs)
    return json.dumps(obj_dict, indent=indent, sort_keys=sort_keys)


def from_json(stream, cls=None, object_pairs_hook=OrderedDict, **extras):
    """
    Convert a JSON string or stream into specified class.
    """
    stream = stream.read() if hasattr(stream, 'read') else stream
    json_dict = json.loads(stream, object_pairs_hook=object_pairs_hook)
    if extras:
        json_dict.update(extras)  # pragma: no cover
    return to_model(cls, json_dict) if cls else json_dict



import related
import pytest


@related.immutable
class Child(object):
    name = related.StringField()
    values = related.SequenceField(str, required=False)


@related.immutable
class Root(object):
    children = related.MappingField(Child, 'name', required=False)
    another = related.MappingField(Child, 'name', default={})


def test_empty_mapping():
    root = related.from_yaml(INPUT_YAML, Root)
    assert related.to_yaml(root).strip() == OUTPUT_YAML


def test_null_mapping():
    root = related.from_yaml(OUTPUT_YAML, Root)
    assert related.to_yaml(root).strip() == OUTPUT_YAML


def test_invalid_mapping():
    with pytest.raises(TypeError):
        related.from_yaml(INVALID_YAML, Root)


INPUT_YAML = "children:"
OUTPUT_YAML = "children: {}\nanother: {}"
INVALID_YAML = "children: 5"



import related


@related.immutable
class SharedModel(object):
    name = related.StringField()
    some_date = related.DateField()


@related.immutable
class InputModel(object):
    lower_case = related.StringField()
    positive_number = related.IntegerField()
    shared = related.MappingField(SharedModel, "name")


@related.immutable
class OutputModel(object):
    upper_case = related.StringField()
    negative_number = related.IntegerField()
    shared = related.MappingField(SharedModel, "name")



from datetime import date, datetime
from uuid import UUID

from future.moves.urllib.parse import ParseResult
from pytest import fixture, raises
from six import string_types

from related import (to_dict, to_yaml, from_yaml,
                     to_json, from_json)

from .models import Company


@fixture
def company():
    return Company(
        name='Acme Inc.',
        url='http://www.acme-inc-website.net/',
        email='info@acme-inc.net',
        meta=dict(key1="value1", key2="value2"),
        nicknames=("Acme", u"Acme", 123),
        temperature=80.4,
        guess=[1, 4, 6, 8, 9, 14, 27, 45],
        established="1/2/1903",
        closed=datetime.strptime('1904-05-06', '%Y-%m-%d'),  # pass datetime
    )


def test_field_getters(company):
    assert company.name == 'Acme Inc.'
    assert isinstance(company.uuid, UUID)
    assert isinstance(company.url, ParseResult)
    assert company.url.geturl() == 'http://www.acme-inc-website.net/'
    assert company.url.netloc == 'www.acme-inc-website.net'
    assert company.url.scheme == 'http'
    assert company.nicknames == ["Acme", "Acme", "123"]
    assert company.guess == {1, 4, 6, 8, 9, 14, 27, 45}
    assert sorted(list(company.guess)) == [1, 4, 6, 8, 9, 14, 27, 45]
    assert company.temperature == 80.4
    assert company.established == date(1903, 1, 2)
    assert company.closed == date(1904, 5, 6)


def test_dictionary(company):
    company_dict = to_dict(company)
    assert company_dict['name'] == 'Acme Inc.'
    assert isinstance(company_dict['uuid'], string_types)
    assert isinstance(company_dict['url'], string_types)
    assert company_dict['established'] == "01/02/1903"
    assert company_dict['closed'] == "1904-05-06"


def test_yaml(company):
    company_yaml = to_yaml(to_dict(company))
    assert ("uuid: %s" % company.uuid) in company_yaml
    assert ("url: %s" % company.url.geturl()) in company_yaml


def test_roundtrip_yaml(company):
    new_company = from_yaml(to_yaml(company), Company)
    assert new_company == company


def test_roundtrip_json(company):
    new_company = from_json(to_json(company), Company)
    assert new_company == company


def test_invalid_regex():
    with raises(TypeError):
        Company(name="fail", email="fail")


def test_invalid_boolean(company):
    with raises(TypeError):
        Company(name="fail", is_active="YES!")


def test_uuid_unique():
    assert Company(name='a').uuid != Company(name='b').uuid



from attr import attr, attributes
import re


@attributes(repr=False, slots=True)
class _CompositeValidator(object):
    validators = attr()

    def __call__(self, inst, attr, value):
        for validator in self.validators:
            validator(inst, attr, value)

    def __repr__(self):
        return (
            "<composite validator for validators {!r}>".format(self.validators)
        )


def composite(*validators):
    """A validator that executes each validator passed as arguments.
    """
    return _CompositeValidator(validators)


@attributes(repr=False, slots=True)
class _RegexValidator(object):
    regex = attr()

    def __call__(self, inst, attr, value):
        if not re.match(self.regex, value):
            raise TypeError(
                "'{name}' must match {regex!r} (got {value!r}).".format(
                    name=attr.name, regex=self.regex, value=value), attr,
                self.regex, value)

    def __repr__(self):
        return (
            "<regex validator for {!r}>".format(self.regex)
        )


def regex(match_string):
    """A validator that executes each validator passed as arguments.

    """
    return _RegexValidator(match_string)



# coding=utf-8
from related.types import TypedSequence, TypedMapping, TypedSet, ImmutableDict
from attr.exceptions import FrozenInstanceError
from related.converters import str_if_not_none
from collections import OrderedDict
import pytest


def test_immutable_dict():
    immutable = ImmutableDict(dict(a=1))

    with pytest.raises(FrozenInstanceError):
        del immutable['a']

    assert immutable == dict(a=1)

    with pytest.raises(FrozenInstanceError):
        immutable['b'] = 2

    assert immutable == dict(a=1)

    with pytest.raises(FrozenInstanceError):
        immutable.clear()

    assert immutable == dict(a=1)

    with pytest.raises(FrozenInstanceError):
        immutable.pop('a')

    assert immutable == dict(a=1)

    with pytest.raises(FrozenInstanceError):
        immutable.something = 0

    assert immutable == dict(a=1)

    with pytest.raises(FrozenInstanceError):
        del immutable.something_else

    assert immutable == dict(a=1)


def test_str_if_not_none():
    unicode_value = "Registered Trademark ®"
    assert unicode_value == str_if_not_none(unicode_value)
    assert "1" == str_if_not_none(1)
    assert str_if_not_none(None) is None


def test_sequence():
    lst = ["a", "b", "c"]
    seq = TypedSequence(str, lst)
    assert seq == lst
    assert str(seq) == str(lst)
    assert repr(seq) == repr(lst)
    assert len(seq) == len(lst)

    del seq[1]
    del lst[1]
    assert seq == lst

    seq[1] = "d"
    assert seq != lst

    with pytest.raises(TypeError):
        seq[1] = 4.0


def test_mapping():
    dct = OrderedDict(a=1, b=2, c=3)
    map = TypedMapping(int, dct)
    assert map == dct
    assert str(map) == str(dct)
    assert repr(map) == repr(dct)
    assert len(map) == len(dct)

    del map["b"]
    del dct["b"]
    assert map == dct

    with pytest.raises(TypeError):
        map["d"] = 4.0

    with pytest.raises(TypeError):
        map.add(5)

    map.add(4, 'd')
    dct['d'] = 4
    assert map == dct


def test_set():
    orig = {"a", "b", "c"}
    typed = TypedSet(str, orig)

    assert orig == typed
    assert len(orig) == len(typed)
    assert 'a' in str(typed)
    assert 'a' in repr(typed)

    typed.add("d")
    assert "d" in typed
    assert orig != typed

    typed.discard("d")
    assert "d" not in typed
    assert orig == typed

    with pytest.raises(TypeError):
        typed.add(5)



from enum import Enum, unique
import related


@unique
class Protocol(Enum):
    TCP = "tcp"
    UDP = "udp"


@unique
class Mode(Enum):
    HOST = "host"
    INGRESS = "ingress"


@related.immutable
class Port(object):
    """ https://docs.docker.com/compose/compose-file/#ports """
    _short_form = related.StringField(required=False, repr=False, cmp=False)
    target = related.IntegerField(required=False)
    published = related.IntegerField(required=False)
    protocol = related.ChildField(Protocol, default=Protocol.TCP)
    mode = related.ChildField(Mode, default=Mode.HOST)

    def __attrs_post_init__(self):
        if self._short_form:
            # todo: handle protocol and mode parsing
            published, target = map(int, self._short_form.split(":"))

            # set methods won't work because the class is immutable.
            # workaround: https://github.com/python-attrs/attrs/issues/120
            object.__setattr__(self, "published", published)
            object.__setattr__(self, "target", target)


@related.to_dict.register(Port)
def _(inst, **kwargs):
    if inst._short_form:
        return inst._short_form
    else:
        return related.to_dict.dispatch(object)(inst, **kwargs)


@related.immutable
class Service(object):
    name = related.StringField()
    image = related.StringField(required=False)
    build = related.StringField(required=False)
    ports = related.SequenceField(Port, required=False)
    volumes = related.SequenceField(str, required=False)
    command = related.StringField(required=False)


@related.immutable
class Compose(object):
    version = related.StringField(required=False, default=None)
    services = related.MappingField(Service, "name", required=False)



import related
from .models import Person, RoleModels


def test_set_construction():
    scientists = [Person(first_name="Grace", last_name="Hopper"),
                  Person(first_name="Katherine", last_name="Johnson"),
                  Person(first_name="Katherine", last_name="Johnson")]

    assert len(scientists) == 3

    role_models = RoleModels(scientists=scientists)
    assert len(role_models.scientists) == 2

    assert related.to_yaml(role_models).strip() in (SET_ORDER_1, SET_ORDER_2)


SET_ORDER_1 = """
scientists:
- first_name: Grace
  last_name: Hopper
- first_name: Katherine
  last_name: Johnson
""".strip()


SET_ORDER_2 = """
scientists:
- first_name: Katherine
  last_name: Johnson
- first_name: Grace
  last_name: Hopper
""".strip()


def test_strict_mode():
    p = related.to_model(Person, dict(first_name="Grace", last_name="Hopper"))
    assert p.first_name == "Grace"

    try:
        d = dict(first_name="Grace", last_name="Hopper", extra="Failure!")
        related.to_model(Person, d)
        assert False, "Did not fail."

    except ValueError as e:
        assert e, "Error as expected."



from os.path import join, dirname
from datetime import datetime, time

from decimal import Decimal

from .models import StoreData, DayType
from related import to_json, from_json, to_model

JSON_FILE = join(dirname(__file__), "store-data.json")


def test_store_data_from_json():
    original_json = open(JSON_FILE).read().strip()
    json_dict = from_json(original_json)
    store_data = to_model(StoreData, json_dict)

    assert store_data.name == "Acme store"
    assert store_data.id == 982
    assert store_data.price == Decimal('98237.448')
    assert store_data.data_from == datetime(2017, 12, 18, 0, 0)
    assert store_data.data_to == datetime(2017, 12, 19, 23, 59, 59)
    assert len(store_data.days) == 2

    assert isinstance(store_data.days[0].open_at, time)
    assert store_data.days[0].date == datetime(2017, 12, 18).date()
    assert store_data.days[0].open_at == time(8, 0, 0)
    assert store_data.days[0].closed_on == time(19, 0, 0)
    assert store_data.days[0].customers == int(487)
    assert store_data.days[0].day_type == DayType.NORMAL
    assert store_data.days[0].sales == float(27223.65)

    assert store_data.days[1].date == datetime(2017, 12, 19).date()
    assert store_data.days[1].open_at == time(10, 30, 0)
    assert store_data.days[1].closed_on == time(17, 30, 0)
    assert store_data.days[1].customers == int(192)
    assert store_data.days[1].day_type == DayType.HOLIDAY
    assert store_data.days[1].sales is None

    generated_json = to_json(store_data,
                             suppress_empty_values=True,
                             suppress_map_key_values=True).replace(" ", "")
    # print(original_json)
    # print(generated_json)
    assert original_json.replace(" ", "") == generated_json



from enum import Enum, unique
import related


@unique
class Degree(Enum):
    HIGH_SCHOOL = "High School"
    ASSOCIATES = "Associate's"
    BACHELORS = "Bachelor's"
    MASTERS = "Master's"
    PHD = "Ph.D"
    JD = "J.D."
    MD = "M.D."
    DDS = "D.D.S."
    PHARMD = "Pharm.D."


@related.immutable
class Address(object):
    street = related.StringField()
    city = related.StringField()
    zipcode = related.StringField()
    street_two = related.StringField(required=False)


@related.immutable
class Education(object):
    school = related.StringField()
    degree = related.ChildField(Degree, required=False)
    field_of_study = related.StringField(required=False)
    from_year = related.IntegerField(required=False)
    to_year = related.IntegerField(required=False)


@related.immutable
class Person(object):
    name = related.StringField()
    age = related.IntegerField(required=False)
    address = related.ChildField(Address, required=False, repr=False)
    education = related.SequenceField(Education, required=False, repr=False)



import related


@related.immutable()
class Child(object):
    name = related.StringField(default="!")


@related.immutable()
class Model(object):
    # non-child fields
    sequence_field = related.SequenceField(str, default=set())
    set_field = related.SetField(str, default=[])
    mapping_field = related.MappingField(Child, "name", default={})

    # child fields
    child_list = related.ChildField(list, default=list)
    child_set = related.ChildField(set, default=set)
    child_dict = related.ChildField(dict, default=dict)
    child_obj = related.ChildField(Child, default=Child)


def test_sequence_field_ok():
    obj_1 = Model()

    obj_1.sequence_field.append("a")
    obj_1.sequence_field.append("b")
    obj_1.sequence_field.append("c")
    obj_1.sequence_field.append("c")

    assert len(obj_1.sequence_field) == 4
    assert len(Model().sequence_field) == 0


def test_set_field_ok():
    obj_1 = Model()

    obj_1.set_field.add("a")
    obj_1.set_field.add("b")
    obj_1.set_field.add("c")
    obj_1.set_field.add("c")

    assert len(obj_1.set_field) == 3
    assert len(Model().set_field) == 0


def test_mapping_field_ok():
    obj_1 = Model()

    obj_1.mapping_field.add(Child("a"))
    obj_1.mapping_field.add(Child("b"))
    obj_1.mapping_field.add(Child("c"))
    obj_1.mapping_field.add(Child("c"))

    assert len(obj_1.mapping_field) == 3
    assert len(Model().mapping_field) == 0


def test_child_list_ok():
    obj_1 = Model()

    obj_1.child_list.append(Child("a"))
    obj_1.child_list.append(Child("b"))
    obj_1.child_list.append(Child("c"))
    obj_1.child_list.append(Child("c"))

    assert len(obj_1.child_list) == 4
    assert len(Model().child_list) == 0


def test_child_set_ok():
    obj_1 = Model()

    obj_1.child_set.add(Child("a"))
    obj_1.child_set.add(Child("b"))
    obj_1.child_set.add(Child("c"))
    obj_1.child_set.add(Child("c"))

    assert len(obj_1.child_set) == 3
    assert len(Model().child_set) == 0


def test_child_dict_ok():
    obj_1 = Model()

    obj_1.child_dict['a'] = Child("a")
    obj_1.child_dict['b'] = Child("b")
    obj_1.child_dict['c'] = Child("c")

    assert len(obj_1.child_dict) == 3
    assert len(Model().child_dict) == 0


def test_child_obj_ok():
    obj_1 = Model(child_obj=Child("a"))

    assert obj_1.child_obj == Child("a")
    assert Model().child_obj != obj_1.child_obj
    assert Model().child_obj == Child()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .models import Person, Address, Education, Degree

from related import to_dict, to_yaml, from_yaml

import pytest


def test_person():
    person = Person(name="Bob")
    assert repr(person) == "Person(name='Bob', age=None)"

    person = Person(name="Bill", age=40)
    assert repr(person) == "Person(name='Bill', age=40)"

    person = Person(name="John", age="50")
    assert repr(person) == "Person(name='John', age=50)"

    d = to_dict(person)
    assert d == {'address': None,
                 'education': [],
                 'age': 50,
                 'name': 'John'}

    new_person = Person(**d)
    assert to_dict(new_person) == d
    assert new_person == person


def test_address_yaml_roundtrip():
    address = Address(street="123 Main Street", city="Springfield",
                      zipcode="12345")
    assert repr(address) == "Address(street='123 Main Street', " \
                            "city='Springfield', zipcode='12345', " \
                            "street_two=None)"

    yaml = to_yaml(address)
    new_address = from_yaml(yaml, Address)
    assert new_address == address


def test_person_with_address():
    address = Address(street="123 Main Street", city="Springfield",
                      zipcode="12345")
    assert repr(address) == "Address(street='123 Main Street', " \
                            "city='Springfield', zipcode='12345', " \
                            "street_two=None)"

    # repr = False for address
    person = Person(name="Jim", address=address)
    assert repr(person) == "Person(name='Jim', age=None)"

    # to_dict() => convert model to dictionary
    d = to_dict(person)
    assert d == {'address': {'city': 'Springfield',
                             'street': '123 Main Street',
                             'street_two': None,
                             'zipcode': '12345'},
                 'education': [],
                 'age': None,
                 'name': 'Jim'}

    # now load from dictionary...
    new_person = Person(**d)
    assert to_dict(new_person) == d
    assert new_person == person


def test_enum():
    education = Education(school="My Alma Mater", degree=Degree.MASTERS)

    # to_dict() => convert model to dictionary
    d = to_dict(education)
    assert d == {'school': 'My Alma Mater',
                 'degree': Degree.MASTERS.value,
                 'field_of_study': None,
                 'from_year': None,
                 'to_year': None}

    # now load from dictionary...
    new_education = Education(**d)
    assert to_dict(new_education) == d
    assert new_education == education

    d = to_dict(new_education)
    assert d == {'school': 'My Alma Mater',
                 'degree': Degree.MASTERS.value,
                 'field_of_study': None,
                 'from_year': None,
                 'to_year': None}


def test_person_with_education_sequence():
    person = Person(name="Brainy", education=[
        Education(school="School 2", degree=Degree.MASTERS),
        Education(school="School 1", degree=Degree.BACHELORS),
    ])

    # to_dict() => convert model to dictionary
    d = to_dict(person)
    assert d == {'address': None,
                 'age': None,
                 'education': [
                     {'degree': Degree.MASTERS.value,
                      'field_of_study': None,
                      'from_year': None,
                      'school': 'School 2',
                      'to_year': None},
                     {'degree': Degree.BACHELORS.value,
                      'field_of_study': None,
                      'from_year': None,
                      'school': 'School 1',
                      'to_year': None}],
                 'name': 'Brainy'}

    # now load from dictionary...
    new_person = Person(**d)
    assert new_person.education == person.education
    assert new_person == person


def test_person_with_education_to_yaml_and_back():
    person = Person(name="Brainy", education=[
        Education(school="School 2", degree=Degree.MASTERS),
        Education(school="School 1", degree=Degree.BACHELORS),
    ])

    yaml = to_yaml(person)
    new_person = from_yaml(yaml, Person)
    assert new_person.education == person.education
    assert new_person == person


def test_fail_missing_field():
    with pytest.raises(TypeError):
        Person()


def test_fail_invalid_age_field():
    with pytest.raises(ValueError):
        Person(name="Joe", age="invalid")


def test_fail_invalid_address_class():
    with pytest.raises(TypeError):
        Person(name="Joe", address="My Address")


def test_with_invalid_class_type():
    Education(school="My Alma Mater")
    with pytest.raises(ValueError):
        Education(school="My Alma Mater", degree="Master of Universe")



from decimal import Decimal
from future.moves.urllib.parse import ParseResult
from collections import OrderedDict
from enum import Enum
from uuid import UUID
from datetime import date, datetime, time

from attr._compat import iteritems

from .functions import to_dict
from .types import (
    TypedSequence, TypedMapping, TypedSet, DEFAULT_DATE_FORMAT,
    DEFAULT_DATETIME_FORMAT, DEFAULT_TIME_FORMAT
)


@to_dict.register(list)  # noqa F811
@to_dict.register(set)
@to_dict.register(tuple)
def _(obj, **kwargs):
    suppress_empty_values = kwargs.get("suppress_empty_values", False)
    retain_collection_types = kwargs.get("retain_collection_types", False)

    if not suppress_empty_values or len(obj):
        cf = obj.__class__ if retain_collection_types else list
        return cf([to_dict(i, **kwargs) for i in obj])


@to_dict.register(dict)  # noqa F811
def _(obj, **kwargs):
    suppress_empty_values = kwargs.get("suppress_empty_values", False)
    dict_factory = kwargs.get("dict_factory", OrderedDict)

    items = []
    for kk, vv in iteritems(obj):
        vv = to_dict(vv, **kwargs)
        if (not suppress_empty_values) or (vv is not None):
            items.append((to_dict(kk, **kwargs), vv))

    if not suppress_empty_values or len(items):
        return dict_factory(items)


@to_dict.register(TypedSequence)  # noqa F811
def _(obj, **kwargs):
    return to_dict(obj.list, **kwargs)


@to_dict.register(TypedSet)  # noqa F811
def _(obj, **kwargs):
    return to_dict(obj.set, **kwargs)


@to_dict.register(TypedMapping)  # noqa F811
def _(obj, **kwargs):
    suppress_map_key_values = kwargs.get("suppress_map_key_values", False)
    suppress_empty_values = kwargs.get("suppress_empty_values", False)
    rv = kwargs.get("dict_factory", OrderedDict)()

    items = obj.items()

    for key_value, item in items:
        sub_dict = to_dict(item, **kwargs)
        if suppress_map_key_values:
            sub_dict.pop(obj.key)
        rv[key_value] = sub_dict

    if not suppress_empty_values or len(items):
        return rv


@to_dict.register(Enum)  # noqa F811
def _(obj, **kwargs):
    return obj.value


@to_dict.register(UUID)  # noqa F811
def _(obj, **kwargs):
    return str(obj)


@to_dict.register(ParseResult)  # noqa F811
def _(obj, **kwargs):
    return obj.geturl()


@to_dict.register(date)  # noqa F811
def _(obj, **kwargs):
    formatter = kwargs.get('formatter') or DEFAULT_DATE_FORMAT
    return obj.strftime(formatter)


@to_dict.register(datetime)  # noqa F811
def _(obj, **kwargs):
    formatter = kwargs.get('formatter') or DEFAULT_DATETIME_FORMAT
    return (obj.isoformat() if formatter == "ISO_FORMAT"
            else obj.strftime(formatter))


@to_dict.register(time)  # noqa F811
def _(obj, **kwargs):
    formatter = kwargs.get('formatter') or DEFAULT_TIME_FORMAT
    return obj.strftime(formatter)


@to_dict.register(Decimal)  # noqa F811
def _(obj, **kwargs):
    return str(obj)



import related
import enum
import uuid


EXAMPLE_UUID = uuid.uuid4()


@enum.unique
class DataType(enum.Enum):
    STRING = 'string'
    NUMBER = 'number'
    INTEGER = 'integer'
    BOOLEAN = 'boolean'
    ARRAY = 'array'
    OBJECT = 'object'


@enum.unique
class IntEnum(enum.Enum):
    a = 1
    b = 2
    c = 3


@related.immutable
class MyChild(object):
    my_int = related.IntegerField(key="int")
    my_float = related.FloatField(key="float")
    my_uuid = related.UUIDField(key="uuid")


@related.immutable
class MyModel(object):
    is_for = related.StringField(key="for")
    criss = related.StringField(key="cross")
    cross = related.StringField(key="criss")
    is_not = related.BooleanField(key="not")
    is_list = related.SequenceField(str, key="list")
    is_type = related.ChildField(DataType, key="type")
    is_dict = related.MappingField(MyChild, "int", key="dict", required=False)
    is_enum = related.ChildField(IntEnum, key="enum", required=False)


def test_renamed():
    obj = MyModel(
        is_for="Elise",
        criss="A",
        cross="B",
        is_not=True,
        is_list=["a", "b", "c"],
        is_dict={5: MyChild(my_int=5, my_uuid=EXAMPLE_UUID, my_float=3.14)},
        is_type=DataType.OBJECT,
        is_enum=IntEnum.a,
    )

    d = related.to_dict(obj)

    assert related.to_model(MyModel, d) == obj

    d.pop("dict")

    assert d == {
        "for": "Elise",
        "criss": "B",
        "cross": "A",
        "not": True,
        'list': ['a', 'b', 'c'],
        "type": "object",
        "enum": 1
    }



"""
Serialization decorator allows for the deserialization of input arguments
and serialization of the output. Useful in the creation of API functions that
need to take a nested-dictionary, transform it into a related object model,
maybe make a new model from that, and then return back a dictionary version
of the response.

Note: this code should also work with Python 3.6 type hints like this:

@related.serializer
def my_function(input: InputMode):
    return OutputModel(
        upper_case=input.lower_case.upper(),
        negative_number=input.positive_number * -1,
        shared=input.shared
    )
"""


import related
from .models import InputModel, OutputModel


@related.serializer(input=InputModel)
def my_function(input):
    return OutputModel(
        upper_case=input.lower_case.upper(),
        negative_number=input.positive_number * -1,
        shared=input.shared
    )


def test_serialization():
    result = my_function(
        input=dict(
            lower_case="hello world.",
            positive_number=123,
            shared=dict(
                one=dict(name="one", some_date="2001-01-01"),
                two=dict(name="two", some_date="2002-02-02"),
            )
        )
    )

    assert result == dict(
        upper_case="HELLO WORLD.",
        negative_number=-123,
        shared=dict(
            one=dict(name="one", some_date="2001-01-01"),
            two=dict(name="two", some_date="2002-02-02"),
        )
    )



# -*- coding: utf-8 -*-
from decimal import Decimal
from future.moves.urllib.parse import ParseResult
from attr import attrib, NOTHING
from collections import OrderedDict
from uuid import uuid4, UUID
from datetime import date, datetime, time
from six import string_types

from . import _init_fields, types, converters, validators


def BooleanField(default=NOTHING, required=True, repr=True, cmp=True,
                 key=None):
    """
    Create new bool field on a model.

    :param default: any boolean value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, bool)
    return attrib(default=default, validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def ChildField(cls, default=NOTHING, required=True, repr=True, cmp=True,
               key=None):
    """
    Create new child field on a model.

    :param cls: class (or name) of the model to be related.
    :param default: any object value of type cls
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    converter = converters.to_child_field(cls)
    validator = _init_fields.init_validator(
        required, object if isinstance(cls, str) else cls
    )
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, cmp=cmp, metadata=dict(key=key))


def DateField(formatter=types.DEFAULT_DATE_FORMAT, default=NOTHING,
              required=True, repr=True, cmp=True, key=None):
    """
    Create new date field on a model.

    :param formatter: date formatter string (default: "%Y-%m-%d")
    :param default: any date or string that can be converted to a date value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, date)
    converter = converters.to_date_field(formatter)
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, cmp=cmp,
                  metadata=dict(formatter=formatter, key=key))


def DateTimeField(formatter=types.DEFAULT_DATETIME_FORMAT, default=NOTHING,
                  required=True, repr=True, cmp=True, key=None):
    """
    Create new datetime field on a model.

    :param formatter: datetime formatter string (default: "ISO_FORMAT")
    :param default: any datetime or string that can be converted to a datetime
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, datetime)
    converter = converters.to_datetime_field(formatter)
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, cmp=cmp,
                  metadata=dict(formatter=formatter, key=key))


def TimeField(formatter=types.DEFAULT_TIME_FORMAT, default=NOTHING,
              required=True, repr=True, cmp=True, key=None):
    """
    Create new time field on a model.

    :param formatter: time formatter string (default: "%H:%M:%S")
    :param default: any time or string that can be converted to a time value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, time)
    converter = converters.to_time_field(formatter)
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, cmp=cmp,
                  metadata=dict(formatter=formatter, key=key))


def FloatField(default=NOTHING, required=True, repr=True, cmp=True,
               key=None):
    """
    Create new float field on a model.

    :param default: any float value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, float)
    return attrib(default=default, converter=converters.float_if_not_none,
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def IntegerField(default=NOTHING, required=True, repr=True, cmp=True,
                 key=None):
    """
    Create new int field on a model.

    :param default: any integer value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, int)
    return attrib(default=default, converter=converters.int_if_not_none,
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def MappingField(cls, child_key, default=NOTHING, required=True, repr=False,
                 key=None):
    """
    Create new mapping field on a model.

    :param cls: class (or name) of the model to be related in Sequence.
    :param child_key: key field on the child object to be used as the map key.
    :param default: any mapping type
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, OrderedDict())
    converter = converters.to_mapping_field(cls, child_key)
    validator = _init_fields.init_validator(required, types.TypedMapping)
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, metadata=dict(key=key))


def RegexField(regex, default=NOTHING, required=True, repr=True, cmp=True,
               key=None):
    """
    Create new str field on a model.

    :param regex: regex validation string (e.g. "[^@]+@[^@]+" for email)
    :param default: any string value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, string_types,
                                            validators.regex(regex))
    return attrib(default=default, converter=converters.str_if_not_none,
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def SequenceField(cls, default=NOTHING, required=True, repr=False, key=None):
    """
    Create new sequence field on a model.

    :param cls: class (or name) of the model to be related in Sequence.
    :param default: any TypedSequence or list
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, [])
    converter = converters.to_sequence_field(cls)
    validator = _init_fields.init_validator(required, types.TypedSequence)
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, metadata=dict(key=key))


def SetField(cls, default=NOTHING, required=True, repr=False, key=None):
    """
    Create new set field on a model.

    :param cls: class (or name) of the model to be related in Set.
    :param default: any TypedSet or set
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, set())
    converter = converters.to_set_field(cls)
    validator = _init_fields.init_validator(required, types.TypedSet)
    return attrib(default=default, converter=converter, validator=validator,
                  repr=repr, metadata=dict(key=key))


def StringField(default=NOTHING, required=True, repr=True, cmp=True,
                key=None):
    """
    Create new str field on a model.

    :param default: any string value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, string_types)
    return attrib(default=default, converter=converters.str_if_not_none,
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def URLField(default=NOTHING, required=True, repr=True, cmp=True, key=None):
    """
    Create new UUID field on a model.

    :param default: any value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    cls = ParseResult
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, cls)
    return attrib(default=default, converter=converters.str_to_url,
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def UUIDField(default=NOTHING, required=False, repr=True, cmp=True, key=None):
    """
    Create new UUID field on a model.

    :param default: any value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    cls = UUID
    default = _init_fields.init_default(required, default, uuid4)
    validator = _init_fields.init_validator(required, cls)
    return attrib(default=default, converter=converters.str_to_uuid,
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))


def DecimalField(default=NOTHING, required=True, repr=True, cmp=True,
                 key=None):
    """
    Create new decimal field on a model.

    :param default: any decimal value
    :param bool required: whether or not the object is invalid if not provided.
    :param bool repr: include this field should appear in object's repr.
    :param bool cmp: include this field in generated comparison.
    :param string key: override name of the value when converted to dict.
    """
    default = _init_fields.init_default(required, default, None)
    validator = _init_fields.init_validator(required, Decimal)
    return attrib(default=default, converter=lambda x: Decimal(x),
                  validator=validator, repr=repr, cmp=cmp,
                  metadata=dict(key=key))



import inspect

from attr import attrs

from .functions import to_model, to_dict, is_model


def mutable(maybe_cls=None, strict=False):

    def wrap(cls):
        wrapped = attrs(cls)
        wrapped.__related_strict__ = strict
        return wrapped

    return wrap(maybe_cls) if maybe_cls is not None else wrap


def immutable(maybe_cls=None, strict=False):

    def wrap(cls):
        wrapped = attrs(cls, frozen=True, slots=True)
        wrapped.__related_strict__ = strict
        return wrapped

    return wrap(maybe_cls) if maybe_cls is not None else wrap


def _get_annotation_map(func, **kwargs):
    annotation_map = kwargs.copy()

    try:
        func_sig = inspect.signature(func)
        for index, arg in enumerate(func_sig.parameters.values()):
            if is_model(arg.annotation):
                annotation_map[arg.name] = arg.annotation
    except AttributeError:
        pass

    return annotation_map


def serializer(func=None, **kwargs):

    def _serializer(func):
        # collect map of related models by name (currently not arg index)
        annotation_map = _get_annotation_map(func, **kwargs)

        def wrapper(*args, **kwargs):
            # iterate keys that intersect
            for key in (set(kwargs.keys()) & set(annotation_map.keys())):
                kwargs[key] = to_model(annotation_map[key], kwargs[key])

            # run function getting result
            result = func(*args, **kwargs)

            # return
            return to_dict(result)

        return wrapper

    return _serializer(func) if func is not None else _serializer



import related


@related.immutable
class Service(object):
    name = related.StringField()
    image = related.StringField(required=False)
    build = related.StringField(required=False)
    ports = related.SequenceField(str, required=False)
    volumes = related.SequenceField(str, required=False)
    command = related.StringField(required=False)


@related.immutable
class Compose(object):
    version = related.StringField(required=False, default=None)
    services = related.MappingField(Service, "name", required=False)



import related


@related.mutable
class Company(object):
    name = related.StringField()
    uuid = related.UUIDField()
    email = related.RegexField("[^@]+@[^@]+", required=False)
    is_active = related.BooleanField(required=False)
    url = related.URLField(required=False)
    meta = related.ChildField(dict, required=False)
    nicknames = related.SequenceField(str, required=False)
    temperature = related.FloatField(required=False)
    guess = related.SetField(int, required=False)
    established = related.DateField('%m/%d/%Y', required=False)
    closed = related.DateField(required=False)  # default formatter (%Y-%m-%d)



# -*- coding: utf-8 -*-

from .decorators import (
    mutable,
    immutable,
    serializer,
)

from .types import (
    ImmutableDict,
    TypedSequence,
    TypedMapping,
    TypedSet,
)

from .fields import (
    BooleanField,
    ChildField,
    DateField,
    DateTimeField,
    TimeField,
    FloatField,
    IntegerField,
    MappingField,
    RegexField,
    SequenceField,
    SetField,
    StringField,
    URLField,
    UUIDField,
    DecimalField,
)

from .functions import (
    from_json,
    from_yaml,
    is_model,
    to_dict,
    to_json,
    to_model,
    to_yaml,
)

from . import dispatchers  # noqa F401

__all__ = [
    # decorators.py
    "mutable",
    "immutable",
    "serializer",

    # types.py
    "ImmutableDict",
    "TypedSequence",
    "TypedMapping",
    "TypedSet",

    # fields.py
    "BooleanField",
    "ChildField",
    "DateField",
    "DateTimeField",
    "TimeField",
    "FloatField",
    "IntegerField",
    "MappingField",
    "RegexField",
    "SetField",
    "StringField",
    "SequenceField",
    "URLField",
    "UUIDField",
    "DecimalField",

    # functions.py
    "from_json",
    "from_yaml",
    "is_model",
    "to_dict",
    "to_json",
    "to_model",
    "to_yaml",
]


__author__ = """Ian Maurer"""
__email__ = 'ian@genomoncology.com'
__version__ = '0.7.2'

__uri__ = "http://www.github.com/genomoncology/related"
__copyright__ = "Copyright (c) 2017 genomoncology.com"
__description__ = "Related: Straightforward nested object models in Python"
__doc__ = __description__ + " <" + __uri__ + ">"
__license__ = "MIT"
__title__ = "related"



import related
from .models import Node

original_json = """
{
  "name": "root",
  "node_list": [
    {
      "name": "A"
    },
    {
      "name": "B"
    }
  ],
  "node_child": {
    "name": "C"
  },
  "node_map": {
    "E": {},
    "F": {
      "node_child": {
        "name": "G"
      }
    }
  }
}
"""


def test_self_reference():
    root_node = related.from_json(original_json, Node)
    assert root_node.name == "root"
    assert root_node.node_list[0].name == "A"
    assert root_node.node_list[1].name == "B"
    assert root_node.node_child.name == "C"
    assert len(root_node.node_map) == 2
    assert root_node.node_map["F"].name == "F"
    assert root_node.node_map["F"].node_child.name == "G"



from os.path import join, dirname

from .models import Compose, Port

from related import to_yaml, from_yaml

YML_FILE = join(dirname(__file__), "docker-compose.yml")


def test_compose_from_yml():
    original_yaml = open(YML_FILE).read().strip()
    yml_dict = from_yaml(original_yaml)
    compose = Compose(**yml_dict)

    # basic checks
    assert compose.version == '2'
    assert compose.services['redis'].image == "redis"

    # compare short form vs. long form
    expected = Port(target="5000", published="5000")
    assert expected == compose.services['web'].ports[0]

    # private attributes suppressed
    generated_yaml = to_yaml(compose,
                             suppress_private_attr=True,
                             suppress_empty_values=True,
                             suppress_map_key_values=True).strip()
    assert "_short_form" not in generated_yaml

    # can we make these equal??
    assert original_yaml == generated_yaml



import related

node_cls = "ex08_self_reference.models.Node"


@related.mutable
class Node(object):
    name = related.StringField()
    node_child = related.ChildField(node_cls, required=False)
    node_list = related.SequenceField(node_cls, required=False)
    node_map = related.MappingField(node_cls, "name", required=False)



import related


# removing the frozen gives you (TypeError: unhashable type: 'Person')
@related.immutable(strict=True)
class Person(object):
    first_name = related.StringField()
    last_name = related.StringField()


@related.immutable
class RoleModels(object):
    scientists = related.SetField(Person)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as history_file:
    history = history_file.read()

with open('dev-requirements.txt') as dev_requirements_file:
    tests_require = [r.strip() for r in dev_requirements_file.readlines()]

setup(
    name="related",
    version='0.7.2',
    url="https://github.com/genomoncology/related",

    package_dir={
        '': 'src'
    },

    packages=[
        "related",
    ],

    include_package_data=True,

    install_requires=[
        "attrs",
        "PyYAML",
        "future",
        "singledispatch;python_version<'3.4'",
        "python-dateutil",
    ],

    setup_requires=[
        'pytest-runner',
    ],

    tests_require=tests_require,

    license="MIT license",

    keywords='related object models yaml json dict nested',
    description="Related: Straightforward nested object models in Python",
    long_description="%s\n\n%s" % (readme, history),

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from collections import OrderedDict
from uuid import UUID
from future.moves.urllib.parse import urlparse
from six import string_types, callable
from datetime import datetime
from inspect import isfunction
from dateutil import parser
from importlib import import_module

from .types import TypedSequence, TypedMapping, TypedSet
from .functions import to_model

CHILD_ERROR_MSG = "Failed to convert value ({}) to child object class ({}). " \
                  + "... [Original error message: {}]"


def to_child_field(cls):
    """
    Returns an callable instance that will convert a value to a Child object.

    :param cls: Valid class type of the Child.
    :return: instance of ChildConverter.
    """

    class ChildConverter(object):

        def __init__(self, cls):
            self._cls = cls

        @property
        def cls(self):
            return resolve_class(self._cls)

        def __call__(self, value):
            try:
                # Issue #33: if value is the class and callable, then invoke
                if value == self._cls and callable(value):
                    value = value()

                return to_model(self.cls, value)
            except ValueError as e:
                error_msg = CHILD_ERROR_MSG.format(value, self.cls, str(e))
                raise ValueError(error_msg)

    return ChildConverter(cls)


def to_sequence_field(cls):
    """
    Returns a callable instance that will convert a value to a Sequence.

    :param cls: Valid class type of the items in the Sequence.
    :return: instance of the SequenceConverter.
    """
    class SequenceConverter(object):

        def __init__(self, cls):
            self._cls = cls

        @property
        def cls(self):
            return resolve_class(self._cls)

        def __call__(self, values):
            values = values or []
            args = [to_model(self.cls, value) for value in values]
            return TypedSequence(cls=self.cls, args=args)

    return SequenceConverter(cls)


def to_set_field(cls):
    """
    Returns a callable instance that will convert a value to a Sequence.

    :param cls: Valid class type of the items in the Sequence.
    :return: instance of the SequenceConverter.
    """
    class SetConverter(object):

        def __init__(self, cls):
            self._cls = cls

        @property
        def cls(self):
            return resolve_class(self._cls)

        def __call__(self, values):
            values = values or set()
            args = {to_model(self.cls, value) for value in values}
            return TypedSet(cls=self.cls, args=args)

    return SetConverter(cls)


def to_mapping_field(cls, key):  # pragma: no mccabe
    """
    Returns a callable instance that will convert a value to a Mapping.

    :param cls: Valid class type of the items in the Sequence.
    :param key: Attribute name of the key value in each item of cls instance.
    :return: instance of the MappingConverter.
    """
    class MappingConverter(object):

        def __init__(self, cls, key):
            self._cls = cls
            self.key = key

        @property
        def cls(self):
            return resolve_class(self._cls)

        def __call__(self, values):
            kwargs = OrderedDict()

            if isinstance(values, TypedMapping):
                return values

            if not isinstance(values, (type({}), type(None))):
                raise TypeError("Invalid type : {}".format(type(values)))

            if values:
                for key_value, item in values.items():
                    if isinstance(item, dict):
                        item[self.key] = key_value
                        item = to_model(self.cls, item)
                    kwargs[key_value] = item

            return TypedMapping(cls=self.cls, kwargs=kwargs, key=self.key)

    return MappingConverter(cls, key)


def str_if_not_none(value):
    """
    Returns an str(value) if the value is not None.

    :param value: None or a value that can be converted to a str.
    :return: None or str(value)
    """
    if not(value is None or isinstance(value, string_types)):
        value = str(value)

    return value


def int_if_not_none(value):
    """
    Returns an int(value) if the value is not None.

    :param value: None or a value that can be converted to an int.
    :return: None or int(value)
    """
    return None if value is None else int(value)


def float_if_not_none(value):
    """
    Returns an float(value) if the value is not None.

    :param value: None or a value that can be converted to an float.
    :return: None or float(value)
    """
    return None if value is None else float(value)


def str_to_url(value):
    """
    Returns a UUID(value) if the value provided is a str.

    :param value: str or UUID object
    :return: UUID object
    """
    return urlparse(value) if isinstance(value, string_types) else value


def str_to_uuid(value):
    """
    Returns a UUID(value) if the value provided is a str.

    :param value: str or UUID object
    :return: UUID object
    """
    if isfunction(value):
        value = value()

    return UUID(value) if isinstance(value, string_types) else value


def to_date_field(formatter):
    """
    Returns a callable instance that will convert a string to a Date.

    :param formatter: String that represents data format for parsing.
    :return: instance of the DateConverter.
    """
    class DateConverter(object):

        def __init__(self, formatter):
            self.formatter = formatter

        def __call__(self, value):
            if isinstance(value, string_types):
                value = datetime.strptime(value, self.formatter).date()

            if isinstance(value, datetime):
                value = value.date()

            return value

    return DateConverter(formatter)


def to_datetime_field(formatter):
    """
    Returns a callable instance that will convert a string to a DateTime.

    :param formatter: String that represents data format for parsing.
    :return: instance of the DateTimeConverter.
    """
    class DateTimeConverter(object):

        def __init__(self, formatter):
            self.formatter = formatter

        def __call__(self, value):
            if isinstance(value, string_types):
                value = parser.parse(value)

            return value

    return DateTimeConverter(formatter)


def to_time_field(formatter):
    """
    Returns a callable instance that will convert a string to a Time.

    :param formatter: String that represents data format for parsing.
    :return: instance of the TimeConverter.
    """
    class TimeConverter(object):

        def __init__(self, formatter):
            self.formatter = formatter

        def __call__(self, value):
            if isinstance(value, string_types):
                value = datetime.strptime(value, self.formatter).time()

            return value

    return TimeConverter(formatter)


def resolve_class(cls):
    if isinstance(cls, str):
        module_name, model_name = cls.rsplit(".", 1)
        module = import_module(module_name)
        cls = getattr(module, model_name)
    return cls



from os.path import join, dirname

from .models import Compose
from related import to_yaml, from_yaml, to_model

YML_FILE = join(dirname(__file__), "docker-compose.yml")


def test_compose_from_yml():
    original_yaml = open(YML_FILE).read().strip()
    yml_dict = from_yaml(original_yaml)
    compose = to_model(Compose, yml_dict)

    assert compose.version == '2'
    assert compose.services['web'].ports == ["5000:5000"]
    assert compose.services['redis'].image == "redis"

    generated_yaml = to_yaml(compose,
                             suppress_empty_values=True,
                             suppress_map_key_values=True).strip()

    assert original_yaml == generated_yaml

    yml_dict2 = from_yaml(generated_yaml)

    compose2 = Compose(**yml_dict2)
    assert compose == compose2
    assert original_yaml == generated_yaml


def test_yaml_roundtrip_with_empty_values():
    original_yaml = open(YML_FILE).read().strip()
    yml_dict = from_yaml(original_yaml)
    compose = Compose(**yml_dict)
    generated_yaml = to_yaml(compose, suppress_map_key_values=True,
                             suppress_empty_values=False).strip()
    assert "ports: []" in generated_yaml



import related


@related.immutable
class ImageOptions(object):
    registry = related.URLField()
    email = related.StringField()


def test_image_options():
    options = ImageOptions(
        registry="https://imgur.com/gallery/GAhlfKS", email="black@knight.com"
    )
    assert options.registry
    assert options.email



from enum import Enum, unique
import related


@unique
class DayType(Enum):
    NORMAL = "Normal"
    HOLIDAY = "Holiday"


@related.immutable
class DayData(object):
    date = related.DateField()
    logged_on = related.TimeField("%H:%M")
    open_at = related.TimeField()
    closed_on = related.TimeField()
    customers = related.IntegerField()
    day_type = related.ChildField(DayType)
    sales = related.FloatField(required=False)


@related.immutable
class StoreData(object):
    name = related.StringField()
    id = related.IntegerField()
    created_on = related.DateTimeField("%m/%d/%Y %H:%M:%S")
    data_from = related.DateTimeField()
    data_to = related.DateTimeField()
    days = related.SequenceField(DayData)
    price = related.DecimalField()
